import cyclone.websocket


class WebSocketHandler(cyclone.websocket.WebSocketHandler):
    def _execute(self, *args, **kwargs):
        if not self._closeIfInvalidMethod():
            return

        if not self._closeIfInvalidUpgradeHeader():
            return

        if not self._closeIfInvalidConnectionHeader():
            return

        super(WebSocketHandler, self)._execute(*args, **kwargs)

    def _closeIfInvalidMethod(self):
        if self.request.method != "GET":
            resp = ( "HTTP/1.1 405 Method Not Allowed\r\n"
                     "Allow: GET\r\n"
                     "Connection: Close\r\n"
                     "\r\n"
                   )
            self._writeAndClose(resp)
            return False
        
        return True

    def _writeAndClose(self, resp):
        self.transport.write(cyclone.escape.utf8(resp))
        self.transport.loseConnection()

    def _closeIfInvalidUpgradeHeader(self):
        if self.request.headers.get("Upgrade", "").lower() != "websocket":
            resp = ( "HTTP/1.1 400 Bad Request\r\n"
                    "Connection: Close\r\n"
                    "\r\n"
                    "Can \"Upgrade\" only to \"WebSocket\"."
                   )
            self._writeAndClose(resp)
            return False

        return True

    def _closeIfInvalidConnectionHeader(self):
        headers = self.request.headers
        connection_headers = [ s.strip().lower() for s in \
                                     headers.get("Connection", "").split(',') ]
        if "upgrade" not in connection_headers:
            resp = ( "HTTP/1.1 400 Bad Request\r\n"
                     "Connection: Close\r\n"
                     "\r\n"
                     "\"Connection\" must be \"Upgrade\"."
                   )
            self._writeAndClose(resp)
            return False

        return True




import random
import hashlib
import time

from twisted.python import log
from twisted.internet import reactor
from twisted.internet import task
from twisted.python.constants import NamedConstant, Names

from sockjs.cyclone import proto
from sockjs.cyclone import utils


class SESSION_STATE(Names):
    """
    Constants representing Session States
    """
    CONNECTING = NamedConstant()
    OPEN = NamedConstant()
    CLOSING = NamedConstant()
    CLOSED = NamedConstant()


class BaseSession(object):
    """ Base session implementation class """

    def __init__(self, conn, server):
        """ Base constructor.

        @param conn: Connection class
        @param server: SockJSRouter instance

        """
        self.server = server
        self.stats = server.stats

        self.send_expects_json = False

        self.handler = None
        self.state = SESSION_STATE.CONNECTING

        self.conn_info = None

        self.conn = conn(self)

        self.close_reason = None

    def set_handler(self, handler):
        """ Set transport handler

        @param handler: Handler, should derive from the
                        C{sockjs.cyclone.transports.base.BaseTransportMixin}
        """
        if self.handler is not None:
            raise Exception('Attempted to overwrite BaseSession handler')

        self.handler = handler
        self.transport_name = self.handler.name

        if self.conn_info is None:
            self.conn_info = handler.get_conn_info()
            self.stats.sessionOpened(self.transport_name)

        return True

    def verify_state(self):
        """ Verify if session was not yet opened. If it is, open it and call
        connection's C{connectionMade} """
        if self.state == SESSION_STATE.CONNECTING:
            self.state = SESSION_STATE.OPEN

            self.conn.connectionMade(self.conn_info)

    def remove_handler(self, handler):
        """ Remove active handler from the session

        @param handler: Handler to remove
        """
        # Attempt to remove another handler
        if self.handler != handler:
            raise Exception('Attempted to remove invalid handler')

        self.handler = None

    def close(self, code=3000, message='Go away!'):
        """ Close session or endpoint connection.

        @param code: Closing code

        @param message: Close message
        """
        if self.state != SESSION_STATE.CLOSED:
            try:
                self.conn.connectionLost()
            except Exception as e:
                log.msg("Failed to call connectionLost(): %r." % e)
            finally:
                self.state = SESSION_STATE.CLOSED
                self.close_reason = (code, message)

            # Bump stats
            self.stats.sessionClosed(self.transport_name)

            # If we have active handler, notify that session was closed
            if self.handler is not None:
                self.handler.session_closed()

    def delayed_close(self):
        """ Delayed close - won't close immediately, but on the next reactor
        loop. """
        self.state = SESSION_STATE.CLOSING
        reactor.callLater(0, self.close)

    def get_close_reason(self):
        """ Return last close reason tuple.

        For example:

            if self.session.is_closed:
                code, reason = self.session.get_close_reason()

        """
        if self.close_reason:
            return self.close_reason

        return (3000, 'Go away!')

    @property
    def is_closed(self):
        """ Check if session was closed. """
        return (self.state == SESSION_STATE.CLOSED 
                or self.state == SESSION_STATE.CLOSING)

    def send_message(self, msg, stats=True):
        """ Send or queue outgoing message

        @param msg: Message to send

        @param stats: If set to True, will update statistics after operation
                      completes
        """
        raise NotImplemented()

    def send_jsonified(self, msg, stats=True):
        """ Send or queue outgoing message which was json-encoded before. Used
        by the C{broadcast} method.

        @param msg: JSON-encoded message to send
        @param stats: If set to True, will update statistics after operation
                      completes
        """
        raise NotImplemented()

    def broadcast(self, clients, msg):
        """ Optimized broadcast implementation. Depending on type of the
        session, will json-encode message once and will call either
        C{send_message} or C{send_jsonifed}.

        @param clients: Clients iterable
        @param msg: Message to send

        """
        self.server.broadcast(clients, msg)


class SessionMixin(object):
    """ Represents one session object stored in the session container.
        Derive from this object to store additional data.
    """

    def __init__(self, session_id=None, expiry=None, time_module=time):
        """ Constructor.

        @param session_id: Optional session id. If not provided, will generate
                           new session id.

        @param expiry: Expiration time in seconds. If not provided, will never
                       expire.

        @param time_module: only used for unit testing. A provider for a time()
                            function
        """
        self.time_module = time_module

        self.session_id = session_id or self._random_key()
        self.promoted = None
        self.expiry = expiry

        if self.expiry is not None:
            self.expiry_date = self.time_module.time() + self.expiry

    def _random_key(self):
        """ Return random session key """
        hashstr = '%s%s' % (random.random(), self.time_module.time())
        return hashlib.md5(hashstr).hexdigest()

    def is_alive(self):
        """ Check if session is still alive """
        return self.expiry_date > self.time_module.time()

    def promote(self):
        """ Mark object as alive, so it won't be collected during next
        run of the garbage collector.
        """
        if self.expiry is not None:
            self.promoted = self.time_module.time() + self.expiry

    def on_delete(self, forced):
        """ Triggered when object was expired or deleted. """
        pass

    def __cmp__(self, other):
        return cmp(self.expiry_date, other.expiry_date)

    def __repr__(self):
        return '%f %s %d' % (getattr(self, 'expiry_date', -1),
                             self.session_id,
                             self.promoted or 0)


class Session(BaseSession, SessionMixin):
    """ SockJS session implementation. """

    def __init__(self, conn, server, session_id, expiry=None):
        """ Session constructor.

        @param conn: Default connection class

        @param server: `SockJSRouter` instance

        @param session_id: Session id

        @param expiry: Session expiry time

        """
        # Initialize session
        self.send_queue = utils.SendQueue()
        self.send_expects_json = True

        # Heartbeat related stuff
        self._heartbeat_timer = None
        self._heartbeat_interval = server.settings['heartbeat_delay']

        self._immediate_flush = server.settings['immediate_flush']
        self._pending_flush = False

        self._verify_ip = server.settings['verify_ip']

        SessionMixin.__init__(self, session_id, expiry)
        BaseSession.__init__(self, conn, server)


    # Session callbacks
    def on_delete(self, forced):
        """ Session expiration callback

        @param forced: If session item explicitly deleted, forced will be set to
                       C{True}. If item expired, will be set to C{False}.
        """
        # Do not remove connection if it was not forced and there's a running
        # connection
        if not forced and self.handler is not None and not self.is_closed:
            self.promote()
        else:
            self.close()

    # Add session
    def set_handler(self, handler, start_heartbeat=True):
        """ Set active handler for the session

        @param handler: Associate active cyclone handler with the session

        @param start_heartbeat: Should session start heartbeat immediately
        """
        # Check if session already has associated handler
        if self.handler is not None:
            handler.send_pack(proto.disconnect(2010,
                                               "Another connection still open"))
            return False

        if self._verify_ip and self.conn_info is not None:
            # If IP address doesn't match - refuse connection
            if handler.request.remote_ip != self.conn_info.ip:
                log.msg('Attempted to attach to session %s (%s) from '
                        'different IP (%s)' % ( self.session_id,
                                                self.conn_info.ip,
                                                handler.request.remote_ip
                                              )
                       )

                handler.send_pack(proto.disconnect(2010, 'Attempted to connect '
                                         'to session from different IP'))
                return False

        if ( self.state == SESSION_STATE.CLOSING
             or self.state == SESSION_STATE.CLOSED):
            handler.send_pack(proto.disconnect(*self.get_close_reason()))
            return False

        # Associate handler and promote session
        super(Session, self).set_handler(handler)

        self.promote()

        if start_heartbeat:
            self.start_heartbeat()

        return True

    def verify_state(self):
        """ Verify if session was not yet opened. If it is, open it and call
        connections C{connectionMade} """
        # If we're in CONNECTING state - send 'o' message to the client
        if self.state == SESSION_STATE.CONNECTING:
            self.handler.send_pack(proto.CONNECT)

        # Call parent implementation
        super(Session, self).verify_state()

    def remove_handler(self, handler):
        """ Detach active handler from the session

        @param handler: Handler to remove
        """
        super(Session, self).remove_handler(handler)

        self.promote()
        self.stop_heartbeat()

    def send_message(self, msg, stats=True):
        """ Send or queue outgoing message

        @param msg: Message to send

        @param stats: If set to True, will update statistics after operation
                      completes
        """
        self.send_jsonified(proto.json_encode(msg), stats)

    def send_jsonified(self, msg, stats=True):
        """ Send JSON-encoded message

        @param msg: JSON encoded string to send
        @param stats: If set to True, will update statistics after operation
                      completes
        """
        assert isinstance(msg, basestring), 'Can only send strings'

        if isinstance(msg, unicode):
            msg = msg.encode('utf-8')

        if self._immediate_flush:
            if self.handler and self.send_queue.is_empty():
                # Send message right away
                self.handler.send_pack('a[%s]' % msg)
            else:
                self.send_queue.push(msg)
                self.flush()
        else:
            self.send_queue.push(msg)

            if not self._pending_flush:
                reactor.callLater(0, self.flush)
                self._pending_flush = True

        if stats:
            self.stats.packSent(1)

    def flush(self):
        """ Flush message queue if there's an active connection running """
        self._pending_flush = False

        if self.handler is None:
            return

        if self.send_queue.is_empty():
            return

        self.handler.send_pack('a[%s]' % self.send_queue.get())
        self.send_queue.clear()

    def close(self, code=3000, message='Go away!'):
        """ Close session.

        @param code: Closing code

        @param message: Closing message
        """
        if self.state != SESSION_STATE.CLOSED:
            # Notify handler
            if self.handler is not None:
                self.handler.send_pack(proto.disconnect(code, message))

        super(Session, self).close(code, message)

    # Heartbeats
    def start_heartbeat(self):
        """ Reset hearbeat timer """
        self.stop_heartbeat()

        self._heartbeat_timer = task.LoopingCall(self._heartbeat)
        self._heartbeat_timer.start(self._heartbeat_interval, False)

    def stop_heartbeat(self):
        """ Stop active heartbeat """
        if self._heartbeat_timer is not None:
            self._heartbeat_timer.stop()
            self._heartbeat_timer = None

    def delay_heartbeat(self):
        """ Delay active heartbeat """
        if self._heartbeat_timer is not None:
            self._heartbeat_timer.reset()

    def _heartbeat(self):
        """ Heartbeat callback """
        if self.handler is not None:
            self.handler.send_pack(proto.HEARTBEAT)
        else:
            self.stop_heartbeat()

    def messagesReceived(self, msg_list):
        """ Handle incoming messages

        @param msg_list: Message list to process
        """
        self.stats.packReceived(len(msg_list))

        for msg in msg_list:
            self.conn.messageReceived(msg)


class MultiplexChannelSession(BaseSession):
    def __init__(self, conn, server, base, name):
        super(MultiplexChannelSession, self).__init__(conn, server)

        self.base = base
        self.name = name

    def send_message(self, msg, stats=True):
        if not self.base.is_closed:
            msg = 'msg,%s,%s' % (self.name, msg)
            self.base.session.send_message(msg, stats)

    def messageReceived(self, msg):
        self.conn.messageReceived(msg)

    def close(self, code=3000, message='Go away!'):
        self.base.sendMessage('uns,%s' % self.name)
        self._close(code, message)

    # Non-API version of the close, without sending the close message
    def _close(self, code=3000, message='Go away!'):
        super(MultiplexChannelSession, self).close(code, message)




from sockjs.cyclone.transports import pollingbase


class StreamingTransportBase(pollingbase.PollingTransportBase):
    def initialize(self, server):
        super(StreamingTransportBase, self).initialize(server)

        self.amount_limit = self.server.settings['response_limit']

        # HTTP 1.0 client might send keep-alive
        if (hasattr(self.request, 'connection') 
                and not self.request.supports_http_1_1()):
            self.request.connection.no_keep_alive = True

    def should_finish(self, data_len):
        """ Check if transport should close long running connection after
        sending X bytes to the client.

        @param data_len: Amount of data that was sent
        """
        self.amount_limit -= data_len

        if self.amount_limit <= 0:
            return True

    def session_closed(self):
        """Called by the session when it was closed"""
        self._detach()

        self.safe_finish()




from .websocket import WebSocketTransport
from .jsonp import JSONPTransport, JSONPSendHandler
from .htmlfile import HtmlFileTransport
from .xhr import XhrPollingTransport, XhrSendHandler
from .xhrstreaming import XhrStreamingTransport
from .eventsource import EventSourceTransport
from .rawwebsocket import RawWebSocketTransport




import heapq
import itertools


class SendQueue(object):
    """ Simple send queue. Stores messages and returns a string of all messages.
    """
    def __init__(self, separator=','):
        self.SEPARATOR = separator
        self._queue = []

    def push(self, msg):
        """ pushes a new message in the queue. """
        self._queue.append(msg)

    def get(self):
        """ returns enqueued messages in a single string, separator-joined """
        return self.SEPARATOR.join(self._queue)
    
    def clear(self):
        """ empties the queue. """
        self._queue = []

    def is_empty(self):
        """ check if the queue is empty. """
        return not self._queue


class PriorityQueue(object):
    """ Simplistic priority queue.
    """
    def __init__(self):
        self._queue = []

        self.counter = itertools.count()   # needed to preserve insertion order
                                           # in elements with the same priority

    def push(self, el):
        """ Put a new element in the queue. """
        count = next(self.counter)
        heapq.heappush(self._queue, (el, count))

    def peek(self):
        """ Returns the highest priority elemente from the queue without
        removing it. """
        return self._queue[0][0]
    
    def pop(self):
        """ Remove and returns the highest priority element form the queue. """
        return heapq.heappop(self._queue)[0]

    def is_empty(self):
        """ Checks if the queue is empty. """
        return not self._queue
    
    def __contains__(self, el):
        return el in [ el[0] for el in self._queue ]




from twisted.python import log

from cyclone.web import asynchronous

from sockjs.cyclone import proto
from sockjs.cyclone.transports import pollingbase


class XhrPollingTransport(pollingbase.PollingTransportBase):
    """ xhr-polling transport implementation """
    name = 'xhr'

    @asynchronous
    def post(self, session_id):
        # Start response
        self.preflight()
        self.handle_session_cookie()

        if not self._attach_session(session_id):
            return

        # Might get already detached because connection was closed in
        # connectionMade
        if not self.session:
            return

        if self.session.send_queue.is_empty():
            self.session.start_heartbeat()
        else:
            self.session.flush()

    def connectionLost(self, reason):
        self.session.delayed_close()

    def send_pack(self, message):
        self.set_header('Content-Type', 'application/javascript; charset=UTF-8')
        self.set_header('Content-Length', len(message) + 1)
        self.write(message + '\n')

        self._detach()

        self.safe_finish()


class XhrSendHandler(pollingbase.PollingTransportBase):
    def post(self, session_id):
        self.preflight()
        self.handle_session_cookie()

        session = self._get_session(session_id)

        if session is None:
            self.set_status(404)
            return

        #data = self.request.body.decode('utf-8')
        data = self.request.body
        if not data:
            self.write("Payload expected.")
            self.set_status(500)
            return

        try:
            messages = proto.json_decode(data)
        except:
            # TODO: Proper error handling
            self.write("Broken JSON encoding.")
            self.set_status(500)
            return

        try:
            session.messagesReceived(messages)
        except Exception:
            log.msg('XHR incoming')
            session.close()

            self.set_status(500)
            return

        self.set_status(204)
        self.set_header('Content-Type', 'text/plain; charset=UTF-8')




#import socket

from twisted.python import log

from sockjs.cyclone import proto
from sockjs.cyclone.transports import base
from sockjs.cyclone import websocket


class WebSocketTransport(websocket.WebSocketHandler, base.BaseTransportMixin):
    """ Websocket transport """
    name = 'websocket'

    def initialize(self, server):
        self.server = server
        self.session = None

    def connectionMade(self, session_id):
        # Stats
        self.server.stats.connectionOpened()

        # Disable nagle if required
        if self.server.settings['disable_nagle']:
           self.transport.setTcpNoDelay(True)

        # Handle session
        self.session = self.server.create_session(session_id, register=False)

        if not self.session.set_handler(self):
            self.transport.loseConnection()
            return

        self.session.verify_state()

        if self.session:
            self.session.flush()

    def _detach(self):
        if self.session is not None:
            self.session.remove_handler(self)
            self.session = None

    def messageReceived(self, message):
        # Ignore empty messages
        if not message or not self.session:
            return

        try:
            msg = proto.json_decode(message)

            if isinstance(msg, list):
                self.session.messagesReceived(msg)
            else:
                self.session.messagesReceived((msg,))
        except Exception as e:
            log.msg('WebSocketTransport.messageReceived: %r' % e)

            # Close session on exception
            #self.session.close()

            # Close running connection
            self.transport.loseConnection()

    def connectionLost(self, reason):
        # Close session if websocket connection was closed
        if self.session is not None:
            # Stats
            self.server.stats.connectionClosed()

            # Detach before closing session
            session = self.session
            self._detach()
            session.close()

    def send_pack(self, message):
        self.sendMessage(message)

    def session_closed(self):
        # If session was closed by the application, terminate websocket
        # connection as well.
        self.transport.loseConnection()
        self._detach()

    # Websocket overrides
    # TODO: no use in cyclone
    def allow_draft76(self):
        return True

    # TODO: no use in cyclone
    def auto_decode(self):
        return False




from twisted.internet import task, reactor

from sockjs.cyclone import transports 
from sockjs.cyclone import session
from sockjs.cyclone import sessioncontainer
from sockjs.cyclone import proto
from sockjs.cyclone import static
from sockjs.cyclone import stats


DEFAULT_SETTINGS = {
    # Sessions check interval in seconds
    'session_check_interval': 1,
    # Session expiration in seconds
    'disconnect_delay': 5,
    # Heartbeat time in seconds. Do not change this value unless
    # you absolutely sure that new value will work.
    'heartbeat_delay': 25,
    # Enabled protocols
    'disabled_transports': [],
    # SockJS location
    'sockjs_url': 'http://cdn.sockjs.org/sockjs-0.3.min.js',
    # Max response body size
    'response_limit': 128 * 1024,
    # Enable or disable JSESSIONID cookie handling
    'jsessionid': True,
    # Should sockjs-cyclone flush messages immediately or queue then and
    # flush on next reactor cycle
    'immediate_flush': True,
    # Enable or disable Nagle for persistent transports
    'disable_nagle': True,
    # Enable IP checks for polling transports. If enabled, all subsequent
    # polling calls should be from the same IP address.
    'verify_ip': True
    }

GLOBAL_HANDLERS = [
    ('xhr_send', transports.XhrSendHandler),
    ('jsonp_send', transports.JSONPSendHandler)
]

TRANSPORTS = {
    'websocket': transports.WebSocketTransport,
    'xhr': transports.XhrPollingTransport,
    'xhr_streaming': transports.XhrStreamingTransport,
    'jsonp': transports.JSONPTransport,
    'eventsource': transports.EventSourceTransport,
    'htmlfile': transports.HtmlFileTransport
}

STATIC_HANDLERS = {
    '/chunking_test': static.ChunkingTestHandler,
    '/info': static.InfoHandler,
    '/iframe[0-9-.a-z_]*.html': static.IFrameHandler,
    '/websocket': transports.RawWebSocketTransport,
    '/?': static.GreetingsHandler
}



class SockJSRouter(object):
    def __init__(self, connection, prefix='', user_settings=dict()):
        self._connection = connection

        self.settings = DEFAULT_SETTINGS.copy()
        if user_settings:
            self.settings.update(user_settings)

        disabled_transports = self.settings['disabled_transports']
        self.websockets_enabled = 'websocket' not in disabled_transports

        self.cookie_needed = self.settings['jsessionid']

        self._initialize_sessions()

        self._initialize_stats()

        self._initialize_urls(prefix)

    def _initialize_urls(self, prefix):
        base = prefix + r'/[^/.]+/(?P<session_id>[^/.]+)'

        kwargs = dict(server=self)

        self._transport_urls = []
        for handler in GLOBAL_HANDLERS:
            url = '%s/%s$' % (base, handler[0])
            reqhandler = handler[1]
            self._transport_urls.append( (url, reqhandler, kwargs) )

        for name, transport in TRANSPORTS.iteritems():
            if name in self.settings['disabled_transports']:
                continue

            # Only version 1 is supported
            url = r'%s/%s$' % (base, name)
            self._transport_urls.append((url, transport, kwargs))

        # Generate static URLs
        for name, handler in STATIC_HANDLERS.iteritems():
            url = '%s%s' % (prefix, name)
            self._transport_urls.append((url, handler, kwargs))

    def _initialize_sessions(self):
        self._sessions = sessioncontainer.SessionContainer()

        check_interval = self.settings['session_check_interval']

        self._sessions_cleanup = task.LoopingCall(self._sessions.expire)
        self._sessions_cleanup.start(check_interval)

        reactor.addSystemEventTrigger('before', 'shutdown',
                                      self._sessions_cleanup.stop)

    def _initialize_stats(self):
        self.stats = stats.StatsCollector()

    def get_connection_class(self):
        """ Return associated connection class """
        return self._connection

    @property
    def urls(self):
        return self._transport_urls

    def apply_routes(self, routes):
        """ Feed list of the URLs to the routes list. Returns list """
        routes.extend(self._transport_urls)
        return routes

    def create_session(self, session_id, register=True):
        """ Creates new session object and returns it.

        @param request: Request that created the session. Will be used to get
                        query string parameters and cookies

        @param register: Should be session registered in a storage. Websockets
                         don't need it.
        """
        s = session.Session(self._connection, self, session_id,
                            self.settings.get('disconnect_delay'))

        if register:
            self._sessions.add(s)

        return s

    def get_session(self, session_id):
        """ Get session by session id

        @param session_id: Session id
        """
        return self._sessions.get(session_id)

    # Broadcast helper
    def broadcast(self, clients, msg):
        """ Optimized C{broadcast} implementation. Depending on type of the
        session, will json-encode message once and will call either
        C{send_message} or C{send_jsonifed}.

        @param clients: Clients iterable

        @param msg: Message to send
        """
        json_msg = None

        count = 0

        for c in clients:
            sess = c.session
            if not sess.is_closed:
                if sess.send_expects_json:
                    if json_msg is None:
                        json_msg = proto.json_encode(msg)
                    sess.send_jsonified(json_msg, False)
                else:
                    sess.send_message(msg, False)

                count += 1

        self.stats.packSent(count)




from twisted.python import log

from sockjs.cyclone.transports import base
from sockjs.cyclone.session import MultiplexChannelSession


class ConnectionInfo(object):
    """ Connection information object.

    Will be passed to the C{connectionMade} handler of your connection class.

    Has few properties:

    @cvar ip: Caller IP address

    @cvar cookies: Collection of cookies

    @cvar arguments: Collection of the query string arguments
    """
    def __init__(self, ip, cookies, arguments):
        self.ip = ip
        self.cookies = cookies
        self.arguments = arguments

    def get_argument(self, name):
        """ Return single argument by name """
        val = self.arguments.get(name)
        if val:
            return val[0]
        return None

    def get_cookie(self, name):
        """ Return single cookie by its name """
        return self.cookies.get(name)


class SockJSConnection(object):
    def __init__(self, session):
        """ Connection constructor.

        @param session: Associated session
        """
        self.session = session

    # Public API
    def connectionMade(self, request):
        """ Default connectionMade() handler.

        Override when you need to do some initialization or request validation.
        If you return False, connection will be rejected.

        You can also throw cyclone HTTPError to close connection.

        @param request: C{ConnectionInfo} object which contains caller IP
                        address, query string parameters and cookies associated
                        with this request (if any).
        @type request: C{ConnectionInfo}
        """
        pass

    def messageReceived(self, message):
        """ Default messageReceived handler. Must be overridden in your
        application """
        raise NotImplementedError()

    def connectionLost(self):
        """ Default connectionLost handler. """
        pass

    def sendMessage(self, message):
        """ Send message to the client.

        @param message: Message to send.
        """
        if not self.is_closed:
            self.session.send_message(message)

    def broadcast(self, clients, message):
        """ Broadcast message to the one or more clients.
        Use this method if you want to send same message to lots of clients, as
        it contains several optimizations and will work fast than just having
        loop in your code.

        @param clients: Clients iterable
        @param message: Message to send
        """
        self.session.broadcast(clients, message)

    def close(self):
        self.session.close()

    @property
    def is_closed(self):
        """ Check if connection was closed """
        return self.session.is_closed


class MultiplexConnection(SockJSConnection):
    channels = {}

    def _messageSplit(self, message):
        parts = message.split(',', 2)
        if len(parts) == 3:
            return parts[0], parts[1], parts[2]
        elif len(parts) == 2:
            return parts[0], parts[1], None
        else:
            raise ValueError

    def connectionMade(self, info):
        self.endpoints = dict()
        self.handler = base.MultiplexTransport(self.session.conn_info)
    
    def messageReceived(self, message):
        try:
            msgtype, topic, payload = self._messageSplit(message)
        except ValueError:
            log.msg('invalid message received <%s>' % message)
            return

        if topic not in self.channels:
            return

        if topic in self.endpoints:
            session = self.endpoints[topic]

            if msgtype == 'uns':
                del self.endpoints[topic]
                session._close()
            elif msgtype == 'msg':
                if payload:
                    session.messageReceived(payload)
        else:
            if msgtype == 'sub':
                session = MultiplexChannelSession(self.channels[topic],
                                         self.session.server,
                                         self,
                                         topic)
                session.set_handler(self.handler)
                session.verify_state()

                self.endpoints[topic] = session

    def connectionLost(self):
        for name, session in self.endpoints.iteritems():
            session._close()

    @classmethod
    def create(cls, **connections):
        channels = dict(channels=connections)
        conn = type(cls.__name__, (cls,), channels) 
        return conn




from __future__ import division

from collections import deque

from twisted.internet import task, reactor


class MovingAverage(object):
    """ Moving average class implementation """
    def __init__(self, period=10):
        """Constructor.

        @param period: Moving window size. Average will be calculated from the
                       data in the window.
        """
        self.period = period
        self.stream = deque()
        self.sum = 0
        self.accumulator = 0
        self.last_average = 0

    def add(self, n):
        """ Add value to the current accumulator

        @param n: Value to add
        """
        self.accumulator += n

    def flush(self):
        """ Add accumulator to the moving average queue and reset it. For
        example, called by the StatsCollector once per second to calculate
        per-second average.
        """
        n = self.accumulator
        self.accumulator = 0

        stream = self.stream
        stream.append(n)
        self.sum += n

        streamlen = len(stream)

        if streamlen > self.period:
            self.sum -= stream.popleft()
            streamlen -= 1

        if streamlen == 0:
            self.last_average = 0
        else:
            self.last_average = self.sum / streamlen


class StatsCollector(object):
    def __init__(self):
        # Sessions
        self.sess_active = 0

        # Avoid circular reference
        self.sess_transports = dict()

        # Connections
        self.conn_active = 0
        self.conn_ps = MovingAverage()

        # Packets
        self.pack_sent_ps = MovingAverage()
        self.pack_recv_ps = MovingAverage()

        self._callback = task.LoopingCall(self._update)
        self._callback.start(1)

        reactor.addSystemEventTrigger('before', 'shutdown', self._callback.stop)

    def _update(self):
        self.conn_ps.flush()

        self.pack_sent_ps.flush()
        self.pack_recv_ps.flush()

    def dump(self):
        """Return dictionary with current statistical information"""
        data = dict( # Sessions
                     sessions_active=self.sess_active,

                     # Connections
                     connections_active=self.conn_active,
                     connections_ps=self.conn_ps.last_average,

                     # Packets
                     packets_sent_ps=self.pack_sent_ps.last_average,
                     packets_recv_ps=self.pack_recv_ps.last_average
        )

        for k, v in self.sess_transports.iteritems():
            data['transp_' + k] = v

        return data

    # Various event callbacks
    def sessionOpened(self, transport):
        self.sess_active += 1

        if transport not in self.sess_transports:
            self.sess_transports[transport] = 0

        self.sess_transports[transport] += 1

    def sessionClosed(self, transport):
        self.sess_active -= 1
        self.sess_transports[transport] -= 1

    def connectionOpened(self):
        self.conn_active += 1
        self.conn_ps.add(1)

    def connectionClosed(self):
        self.conn_active -= 1

    def packSent(self, num):
        self.pack_sent_ps.add(num)

    def packReceived(self, num):
        self.pack_recv_ps.add(num)




""" Various base http handlers """

import datetime

from cyclone.web import asynchronous, RequestHandler


class BaseHandler(RequestHandler):
    """ Base request handler with set of helpers."""

    CACHE_TIME = 31536000


    def initialize(self, server):
        """ Initialize request

        @param server: SockJSRouter instance.

        """
        self.server = server
        self.logged = False

    # Statistics
    def prepare(self):
        """ Increment connection count """
        self.logged = True
        self.server.stats.connectionOpened()

    def _log_disconnect(self):
        """ Decrement connection count """
        if self.logged:
            self.server.stats.connectionClosed()
            self.logged = False

    def finish(self):
        """ Cyclone C{finish} handler """
        self._log_disconnect()

        super(BaseHandler, self).finish()

    def on_connection_close(self, reason):
        self._log_disconnect()

    # Various helpers
    def enable_cache(self):
        """ Enable client-side caching for the current request """
        self.set_header('Cache-Control', 'max-age=%d, public' % self.CACHE_TIME)

        now = datetime.datetime.now()
        expires = now + datetime.timedelta(seconds=self.CACHE_TIME)
        self.set_header('Expires', expires.strftime('%a, %d %b %Y %H:%M:%S'))

        self.set_header('access-control-max-age', self.CACHE_TIME)

    def disable_cache(self):
        """ Disable client-side cache for the current request """
        self.set_header('Cache-Control',
                        'no-store, no-cache, must-revalidate, max-age=0')

    def handle_session_cookie(self):
        """ Handle JSESSIONID cookie logic """
        # If JSESSIONID support is disabled in the settings, ignore cookie logic
        if not self.server.settings['jsessionid']:
            return

        cookie = self.cookies.get('JSESSIONID')

        if not cookie:
            cv = 'dummy'
        else:
            cv = cookie.value

        self.set_cookie('JSESSIONID', cv)

    def safe_finish(self):
        """ Finish session safely (do not raise if already finished by a dropped
        connection). """
        try:
            self.finish()
        except Exception:
            pass


class PreflightHandler(BaseHandler):
    """ CORS preflight handler """

    @asynchronous
    def options(self, *args, **kwargs):
        """ XHR cross-domain OPTIONS handler """
        self.enable_cache()
        self.handle_session_cookie()
        self.preflight()

        if self.verify_origin():
            allowed_methods = getattr(self, 'access_methods', 'OPTIONS, POST')
            self.set_header('Access-Control-Allow-Methods', allowed_methods)
            self.set_header('Allow', allowed_methods)

            self.set_status(204)
        else:
            # Set forbidden
            self.set_status(403)

        self.finish()

    def preflight(self):
        """ Handles request authentication """
        origin = self.request.headers.get('Origin', '*')

        # 'null' may be sent by the browser when the sockjs client is hosted
        # from a file:// url.
        # In such case, respond with '*' to 'null' origin, as per sockjs
        # protocol definition
        if origin == 'null':
            origin = '*'

        self.set_header('Access-Control-Allow-Origin', origin)

        headers = self.request.headers.get('Access-Control-Request-Headers')
        if headers:
            self.set_header('Access-Control-Allow-Headers', headers)

        self.set_header('Access-Control-Allow-Credentials', 'true')

    def verify_origin(self):
        """ Verify if request can be served """
        # TODO: Verify origin
        return True




from cyclone.web import asynchronous

from sockjs.cyclone.transports import streamingbase


class EventSourceTransport(streamingbase.StreamingTransportBase):
    name = 'eventsource'

    @asynchronous
    def get(self, session_id):
        # Start response
        self.preflight()
        self.handle_session_cookie()
        self.disable_cache()

        self.set_header('Content-Type', 'text/event-stream; charset=UTF-8')
        self.write('\r\n')
        self.flush()

        if not self._attach_session(session_id):
            self.finish()
            return

        if self.session:
            self.session.flush()

    def connectionLost(self, reason):
        self.session.delayed_close()
        self._detach()

    def send_pack(self, message):
        msg = 'data: %s\r\n\r\n' % message

        self.write(msg)
        self.flush()

        # Close connection based on amount of data transferred
        if self.should_finish(len(msg)):
            self._detach()
            self.safe_finish()




import urllib

from cyclone.web import asynchronous

from twisted.python import log

from sockjs.cyclone import proto
from sockjs.cyclone.transports import pollingbase


class JSONPTransport(pollingbase.PollingTransportBase):
    name = 'jsonp'

    @asynchronous
    def get(self, session_id):                                                  
        # Start response                                                        
        self.handle_session_cookie()                                            
        self.disable_cache()                                                    
                                                                                
        # Grab callback parameter                                               
        self.callback = self.get_argument('c', None)                            
        if not self.callback:                                                   
            self.write('"callback" parameter required')                         
            self.set_status(500)                                                
            self.finish()                                                       
            return                                                              

        # Get or create session without starting heartbeat
        if not self._attach_session(session_id):
            return

        # Might get already detached because connection was closed in
        # connectionMade
        if not self.session:
            return

        if self.session.send_queue.is_empty():
            self.session.start_heartbeat()
        else:
            self.session.flush()

    def connectionLost(self, reason):
        self.session.delayed_close()

    def send_pack(self, message):
        # TODO: Just escape
        msg = '%s(%s);\r\n' % (self.callback, proto.json_encode(message))

        self.set_header('Content-Type',
                        'application/javascript; charset=UTF-8')
        self.set_header('Content-Length', len(msg))

        # FIXME
        self.set_header('Etag', 'dummy')

        self.write(msg)

        self._detach()

        self.safe_finish()


class JSONPSendHandler(pollingbase.PollingTransportBase):
    def post(self, session_id):
        self.preflight()
        self.handle_session_cookie()

        session = self._get_session(session_id)

        if session is None:
            self.set_status(404)
            return

        #data = self.request.body.decode('utf-8')
        data = self.request.body

        ctype = self.request.headers.get('Content-Type', '').lower()
        if ctype == 'application/x-www-form-urlencoded':
            if not data.startswith('d='):
                log.msg('jsonp_send: Invalid payload.')

                self.write("Payload expected.")
                self.set_status(500)
                return

            data = urllib.unquote_plus(data[2:])

        if not data:
            log.msg('jsonp_send: Payload expected.')

            self.write("Payload expected.")
            self.set_status(500)
            return

        try:
            messages = proto.json_decode(data)
        except:
            # TODO: Proper error handling
            log.msg('jsonp_send: Invalid json encoding')

            self.write("Broken JSON encoding.")
            self.set_status(500)
            return

        try:
            session.messagesReceived(messages)
        except Exception:
            log.msg('jsonp_send: messagesReceived() failed')

            session.close()

            self.write('Message handler failed.')
            self.set_status(500)
            return

        self.write('ok')
        self.set_header('Content-Type', 'text/plain; charset=UTF-8')
        self.set_status(200)




from twisted.python import log

from sockjs.cyclone import session
from sockjs.cyclone.transports import base
from sockjs.cyclone import websocket



class RawSession(session.BaseSession):
    """ Raw session without any sockjs protocol encoding/decoding.
        Simply works as a proxy between C{SockJSConnection} class and
        C{RawWebSocketTransport}. """
    def sendMessage(self, msg):
        self.handler.send_pack(msg)

    def messageReceived(self, msg):
        self.conn.messageReceived(msg)


class RawWebSocketTransport(websocket.WebSocketHandler, base.BaseTransportMixin):
    """ Raw Websocket transport """

    name = 'rawwebsocket'

    def initialize(self, server):
        self.server = server
        self.session = None

    def open(self):
        # Stats
        self.server.stats.connectionOpened()

        # Disable nagle if required
        if self.server.settings['disable_nagle']:
            self.transport.setTcpNoDelay(True)

        # Create and attach to session
        self.session = RawSession(self.server.get_connection_class(), self.server)
        self.session.set_handler(self)
        self.session.verify_state()

    def _detach(self):
        if self.session is not None:
            self.session.remove_handler(self)
            self.session = None

    def messageReceived(self, message):
        # Ignore empty messages
        if not message or not self.session:
            return

        try:
            self.session.messageReceived(message)
        except Exception as e:
            log.msg('RawWebSocketTransport.messageReceived: %r' % e)

            # Close running connection
            self._detach()
            self.transport.loseConnection()

    def connectionLost(self, reason):
        # Close session if websocket connection was closed
        if self.session is not None:
            # Stats
            self.server.stats.connectionClosed()

            session = self.session
            self._detach()
            session.close()

    def send_pack(self, message):
        # Send message
        self.write_message(message)

    def session_closed(self):
        self.close()

    # Websocket overrides
    # TODO: unused in cyclone
    def allow_draft76(self):
        return True

